#include "global.hpp"
#define RES PIN_DTR
void enter_bootmode() {
	data[0] = 0;
	FT_Write(handle, data, 1, &written);
	FT_Read(handle, data, 1, &written);
	usleep(60*1000);
}

void exit_bootmode() {
	data[0] = RES;
	FT_Write(handle, data, 1, &written);
	FT_Read(handle, data, 1, &written);
}

FT_STATUS write_target(uint32_t dt) {
	bytes = dt;
	packInArray(data, bytes);
	return FT_Write(handle, data, 64, &written);
}

FT_STATUS write_target2(uint32_t dt1, uint32_t dt2) {
	uint64_t bytes = ((uint64_t)dt1 << 32) | dt2;
	printf("%" PRIX64 "\n", bytes);
	packInArray2(data, bytes);
	return FT_Write(handle, data, 128, &written);
}

FT_STATUS read_target() {
	return FT_Read(handle, data, 64, &written);
}

FT_STATUS read_target2() {
	return FT_Read(handle, data, 128, &written);
}

uint16_t readAt(uint16_t addr) {
	write_target(READ_FLASH_LOW | (addr << 8));
	write_target(READ_FLASH_HIGH | (addr << 8));
	read_target();
	uint64_t dat = expandFromArray(data) & 0xFF;
	read_target();
	dat = dat | ((expandFromArray(data) & 0xFF) << 8);
	return dat;
}

void polling_while_busy() {
	uint32_t callback;
	do {
		write_target(IS_BUSY);
		read_target();
		callback = expandFromArray(data);
	} while( (callback & 0xFF) == 1 );
}

struct DevInfo {
	uint64_t devid;
	const char *devname;
};

struct DevInfo info[] = {
	{0x1E9514, "ATmega328"},
	{0x1E950F, "ATmega328P"},
	{0x1E9108, "ATtiny25"},
	{0x1E9206, "ATtiny45"},
	{0x1E930B, "Attiny85",}
};

const char *search_devid(uint64_t key) {
	for(auto &item: info) {
		if( item.devid == key ) {
			return item.devname;
		}
	}
	return "unknown";
}

void ask_devinfo() {
	write_target(READ_DEVID | 0x0000);
	read_target();
	uint64_t id = expandFromArray(data) & 0xFF;
	printf("DevID: %02" PRIX64, id);

	write_target(READ_DEVID | 0x0100);
	read_target();
	id = (id << 8) | (expandFromArray(data) & 0xFF);
	printf("%02" PRIX32 , expandFromArray(data) & 0xFF);

	write_target(READ_DEVID | 0x0200);
	read_target();
	id = (id << 8) | (expandFromArray(data) & 0xFF);
	printf("%02" PRIX32 " :: %s\n", expandFromArray(data) & 0xFF, search_devid(id));

	write_target(READ_LFUSE);
	read_target();
	printf("lfuse: %02" PRIX32 "\n", expandFromArray(data) & 0xFF);

	write_target(READ_HFUSE);
	read_target();
	printf("hfuse: %02" PRIX32 "\n", expandFromArray(data) & 0xFF);

	write_target(READ_EFUSE);
	read_target();
	printf("efuse: %02" PRIX32 "\n", expandFromArray(data) & 0xFF);
}

static char *nextToken(const char *delim) {
	return strtok(NULL, delim);
}

void flash() {
	static const char *delim = "\n\r";
	std::ifstream ifs(filename);
	if( ifs.fail() ) {
		printf("Failed to open file %s.\n", filename);
		return;
	}

	std::string str((std::istreambuf_iterator<char>(ifs)), std::istreambuf_iterator<char>()); // ファイルまるごと読み込み
	char buf[str.size() + 1];
	strcpy(buf, str.c_str());

	strtok(buf, delim);
	char *tp = buf;

	uint16_t word_p = 0x00, page_p = 0x00;
	printf("Page #%02X: ", page_p);
	while( tp != NULL ) {
		if( tp[7] == '0' && tp[8] == '2' ) {
			tp = nextToken(delim);
			continue;
		}
		int offset = c_num(tp[3]) * 16*16*16 + c_num(tp[4]) * 16*16 + c_num(tp[5]) * 16 + c_num(tp[6]);
		word_p = offset / 2;
		page_p = 0;
		while( word_p >= 64 ) {
			page_p++;
			word_p -= 64;
		}
		int prog_bytes = c_num(tp[1]) * 16 + c_num(tp[2]);
		tp += 9; // レコードなんちゃらはパース済みに付きスキップ
		uint8_t dat_l, dat_h;
		for(int i=0; i<prog_bytes/2; i++){
			dat_l = c_num(tp[0]) * 16 + c_num(tp[1]);
			dat_h = c_num(tp[2]) * 16 + c_num(tp[3]);
			write_target(SET_FLASH_PAGE_LOW | (word_p << 8) | dat_l);
			polling_while_busy();
			read_target();
			write_target(SET_FLASH_PAGE_HIGH | (word_p << 8) | dat_h);
			polling_while_busy();
			read_target();
			write_target(WRITE_FLASH_PAGE | (page_p << 14));
			polling_while_busy();

			word_p++;
			if( word_p == 64 ) {
				word_p = 0;
				page_p++;
				printf("\nPage #%02X: ", page_p);
			}
			tp += 4;
			printf("|");
			fflush(stdout);
			read_target();
		}
		tp = nextToken(delim);
	}
	puts("");
}

/*	プログラミングめも
 *	- AVRはリトルエンディアン(12EF -> EF12の順で1ワード書き込み)
 *	- ATmega328Pの内蔵フラッシュは合計で32KByte(16KWord)
 *	- PC0~7に各64Wordの容量がある
 *	- そのページ内でのバイト指定？？
 *
 *	- AVR910によると、まずページはどうでもいいとして、バイト位置と上位・下位バイトを書き込む
 *	- そんで、フラッシュページ書き込み命令によって実際に書き込むページ番号を指定し書き込み実行、らしい。。。
 */

void write_lfuse(uint8_t val) {
	write_target(WRITE_LFUSE | val);
	read_target();
}

void write_hfuse(uint8_t val) {
	write_target(WRITE_HFUSE | val);
	read_target();
}

void write_efuse(uint8_t val) {
	write_target(WRITE_EFUSE | val);
	read_target();
}
